from transferboost.utils.boost import TBoost
from transferboost.utils import UnsupportedModelError, assure_numpy_array
from sklearn.utils.validation import check_is_fitted


class LGBMTransferLearner(TBoost):
    """Main class to transfer boost a trained xgb model.

    Example of use
    ```python
        import transferboost
        from transferboost.dataset import load_data
        from transferboost.models import LGBTransferLearner

        import lightgbm as lgb

        # Get the data and the targets
        X, y1, y2 = load_data(return_X_y=True)

        lgb_model = lgb.LGBMClassifier(
            max_depth = 3,
            reg_lambda = 0,
            num_leaves=5,
            n_estimators=100,
        )

        lgb_model.fit(X,y1)

        t_lgb_model = LGBTransferLearner(lgb_model)
        t_lgb_model.fit(X,y2)
        t_lgb_model.predict_proba(X)
    ```
    """

    def __init__(self, model, base_score=None, f_obj=None, verbosity=0):
        """Constructor for XGBTransferLearner.

        Args:
            model: XGBClassifier model. Must be fitted.
            base_score (float): base score for xgboost model. The default value of None will set the base_score
                to the one of the fitted model.
            f_obj (str of func): objective function. Must return the gradient and hessian of the desired loss
                function. To be passed to the constructor of TBoost.
                Default is None (it will use the standard binary logloss function).
            verbosity (int): verbosity level. If 0 no outputs printed, if >0 returns warnings
        """
        try:
            import lightgbm as lgb
        except ImportError as e:
            raise ImportError("No lightgbm installed. Please install via pip install lightgbm") from e

        if not isinstance(model, lgb.LGBMClassifier):
            msg = f"{self.__class__.__name__} does not support the model {model.__class__.__name__}"
            raise UnsupportedModelError(msg)
        check_is_fitted(model)
        self.model = model

        self.base_score = base_score

        model_params = model.get_params()

        if f_obj is None:
            f_obj = "logloss"

        super().__init__(model_params=model_params, loss_func=f_obj, base_score=base_score, verbosity=verbosity)

    def __repr__(self):
        """Object representation for class XGBTransferLearner."""
        repr_ = f"XGBTransferLearner with base model\n\t{self.model}\n"

        if self.base_score is not None:
            repr_ += f"base_score = {self.base_score}"

        return repr_

    def fit(self, X, y):
        """Fit the XGBTransferLearner.

        Args:
            X: feature set
            y: target (to transfer learn it too)
        """
        y = assure_numpy_array(y, assure_1d=True)
        if self.base_score is None:
            self.base_score = y.mean()
            self._set_base_score(base_score=self.base_score)
        X_leaves_ix = self.model.predict(X, pred_leaf=True)
        return self._fit(X_leaves_ix, y)

    def predict_proba(self, X, tree_index=-1):
        """Predict the probabilities after transfer learning.

        Args:
            X: features
            tree_index: index of the tree up to which to predict the

        Returns (np.array): array of shape (X.shape[0],2). Transfer-learned predictions
            for class 0 and class 1.
        """
        X_leaves_ixs = self.model.predict(X, pred_leaf=True)

        probas = self._predict_proba(X_leaves_ixs=X_leaves_ixs, tree_index=tree_index)

        return probas



import numpy as np
import pandas as pd


def loss_from_leaves(y_leaf, y_true, loss_func):
    """Calculate the gradients and hessian of the logloss functions.

    Args:
        y_leaf (np.array or pd.Series): output of the leaves of the latest tree
        y_true (np.array or pd.Series): true binary target
        loss_func (function): loss function to be used. Must return a tuple of arrays,
            corresponding to the gradient and hessian of the loss function.

    Returns:
        [type]: [description]
    """
    # Apply the sigmoid transformation to the leaf values (transfrom it to probability)
    prob = _logistic(y_leaf)
    # print(f"Probabilities:\n{prob}")

    return loss_func(prob, y_true)


def logloss(y_pred, y_true):
    """Return the gradient and hessian of the log loss.

    Args:
        y_pred (np.array or pd.Series): predictions for the observations, in the range [0,1]
        y_true (np.array or pd.Series): true label for the observations

    Returns:
        tuple: (np.array, np.array), gradient and hessian for the observations
    """
    if isinstance(y_pred, pd.Series):
        y_pred = y_pred.values
    if isinstance(y_true, pd.Series):
        y_true = y_true.values
    grad = y_true - y_pred
    hess = y_pred * (1.0 - y_pred)
    return grad, hess


def _sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))


def _logistic(x):
    return np.exp(x) / (1.0 + np.exp(x))



import pkgutil
import io
import pandas as pd


def load_data(return_X_y=False, as_frame=False):
    """Loads a simulated data set with two targets, y_1 and y_2.

    y_1 and y_2 are generated
    Args:
        return_X_y:  (bool) If True, returns ``(data, target)`` instead of a dict object.
        as_frame: (bool) give the pandas dataframe instead of X, y matrices (default=False).
    Returns: (pd.DataFrame, dict or tuple) features and target, with as follows:
        - if as_frame is True: returns pd.DataFrame with y as a target
        - return_X_y is True: returns a tuple: (X,y)
        - is both are false (default setting): returns a dictionary where the key `data` contains the features,
        and the key `target` is the target
    """
    file = pkgutil.get_data("transferboost", "data/data.zip")
    df = pd.read_csv(io.BytesIO(file), compression="zip")

    if as_frame:
        return df

    X, y1, y2 = (
        df[[col for col in df.columns if col.startswith("f_")]],
        df["y_1"],
        df["y_2"],
    )
    if return_X_y:
        return X, y1, y2

    return {"data": X, "target_1": y1, "target_2": y2}



class NotFittedError(Exception):
    """Error to be returned when the model is not fitted."""

    def __init__(self, message):
        """Error constructor.

        Args:
            message (string): error message
        """
        self.message = message


class DimensionalityError(Exception):
    """Error to be returned when the array dimensionality is wrong."""

    def __init__(self, message):
        """Error constructor.

        Args:
            message (string): error message
        """
        self.message = message


class UnsupportedModelError(Exception):
    """Error to be Model is not supported by the package."""

    def __init__(self, message):
        """Error constructor.

        Args:
            message (string): error message
        """
        self.message = message



from .exceptions import NotFittedError, DimensionalityError, UnsupportedModelError
from .boost import TBoost
from .utils import assure_numpy_array

__all__ = ["NotFittedError", "DimensionalityError", "UnsupportedModelError", "TBoost", "assure_numpy_array"]



import numpy as np
from functools import reduce

def plot_proba_evol(X,y,model, max_n_trees=500):
    arrays = list()
    for ix in range(1,max_n_trees+1):
        try:
            pred_ix = model.predict_proba(X, num_iteration=ix)[:,1]
        except:
            pred_ix = model.predict_proba(X, ntree_limit=ix)[:,1]
        arrays.append(pred_ix)
    return np.transpose(reduce(lambda x,y: np.vstack((x,y)),arrays))        
        
    
def find_pred_trend(X,y,model, max_n_trees = 500):
    
    proba_trees = plot_proba_evol(X,y,model, max_n_trees=max_n_trees)
    
#     return proba_trees
    
    mean_y_0 = proba_trees[y.values==0].mean(axis=0)
    std_y_0 = proba_trees[y.values==0].std(axis=0)

    mean_y_1 = proba_trees[y.values==1].mean(axis=0)
    std_y_1 = proba_trees[y.values==1].std(axis=0)
    
    return mean_y_0, std_y_0, mean_y_1, std_y_1


import numpy as np
import pandas as pd
import numbers
from .exceptions import DimensionalityError


def check_1d(x):
    """
    Checks whether or not a list, numpy array, pandas dataframe, pandas series are one-dimensional.

    Returns True when check is ok, otherwise throws a `DimensionalityError`.

    Args:
        x: list, numpy array, pandas dataframe, pandas series
    Returns: True or throws `DimensionalityError`
    """
    if isinstance(x, list):
        if any([isinstance(el, list) for el in x]):
            raise DimensionalityError("The input is not 1D")
        else:
            return True
    if isinstance(x, np.ndarray):
        if x.ndim == 1 and all([isinstance(el, numbers.Number) for el in x]):
            return True
        else:
            raise DimensionalityError("The input is not 1D")
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1 and pd.api.types.is_numeric_dtype(x[x.columns[0]]):
            return True
        else:
            raise DimensionalityError("The input is not 1D")
    if isinstance(x, pd.core.series.Series):
        if x.ndim == 1 and pd.api.types.is_numeric_dtype(x):
            return True
        else:
            raise DimensionalityError("The input is not 1D")


def assure_numpy_array(x, assure_1d=False):
    """
    Returns x as numpy array. X can be a list, numpy array, pandas dataframe, pandas series.

    Args:
        x: list, numpy array, pandas dataframe, pandas series
        assure_1d: whether or not to assure that the input x is one-dimensional
    Returns: numpy array
    """
    if assure_1d:
        _ = check_1d(x)
    if isinstance(x, list):
        return np.array(x)
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1:
            return x.values.flatten()
        else:
            return x.values
    if isinstance(x, pd.core.series.Series):
        return x.values


def assure_pandas_df(x, column_names=None):
    """
    Returns x as pandas DataFrame. X can be a list, list of lists, numpy array, pandas DataFrame or pandas Series.

    Args:
        x (list, numpy array, pandas DataFrame, pandas Series): array to be tested

    Returns:
        pandas DataFrame
    """
    if isinstance(x, pd.DataFrame):
        # Check if column_names are passed correctly
        if column_names is not None:
            x.columns = column_names
        return x
    elif any([isinstance(x, np.ndarray), isinstance(x, pd.core.series.Series), isinstance(x, list)]):
        return pd.DataFrame(x, columns=column_names)
    else:
        raise TypeError("Please supply a list, numpy array, pandas Series or pandas DataFrame")


def check_numeric_dtypes(x):
    """
    Checks if all entries in an array are of a data type that can be interpreted as numeric (int, float or bool).

    Args:
        x (np.ndarray or pd.Series, list): array to be checked

    Returns:
        x: unchanged input array

    Raises:
        TypeError: if not all elements are of numeric dtypes
    """
    x = assure_numpy_array(x)
    allowed_types = [bool, int, float]

    for element in np.nditer(x):
        if type(element.item()) not in allowed_types:
            raise TypeError("Please supply an array with only floats, ints or booleans")
    return x



import numpy as np
import warnings

from .utils import assure_numpy_array
from .loss_functions import logloss, _logistic


class TBoost:
    """Base class for Transfer Boosting.

    To be inherited in transferboost.models.xgb or transferboost.models.lgb
    """

    def __init__(self, model_params=None, loss_func="logloss", base_score=None, verbosity=0):
        """Constructor for TBoost.

        Args:
            model_params (dict): contains the model parameters of the boosted model.
            loss_func (str or function): loss function, returns gradient and hessians
                of the loss function. Currently supports only loss_func=='logloss'
            base_score (float): starting probability, None by default.
            verbosity (int): verbosity flag.
        """
        self.model_params = model_params
        if loss_func == "logloss":
            self.f_loss_func = logloss
        else:
            raise NotImplementedError(f"Loss function {loss_func} not supported currently")

        if base_score is not None:
            self._set_base_score(base_score)

        self.verbosity = verbosity

    def _set_base_score(self, base_score):
        """Set the base score.

        Args:
            base_score (float): base score.
        """
        self.base_score = base_score

        # Define the starting vector of leaves for the transferboost
        if not 0 < self.base_score < 1:
            raise ValueError(f"Starting proba must be between 0 and 1. Passed {self.base_score}")
        self.start_odds = np.log(self.base_score / (1 - self.base_score))

    def _fit(self, X_leaves_ixs, y):
        """Recompute the output value of the leaves.

        Args:
            X_leaves_ixs ([type]): [description]
            y ([type]): [description]

        Raises:
            ValueError: [description]
            NotImplementedError: [description]

        Returns:
            tuple (np.array, dict):
                leaves_val_array: numpy array of shape (n_rows, n_trees+1), containing the values for every
                    output of the tree. The first column on the array corresponds to the bias term,
                    and is equalfor every row. The second column corresponds to the first tree...
                leaf_vals_map: dictionary, contains the mapping for every tree leaf with the recomputed
                    leaf value.
        """
        X_leaves_ixs = assure_numpy_array(X_leaves_ixs)
        y = assure_numpy_array(y, assure_1d=True)
        # X = assure_numpy_array(X)

        n_rows = X_leaves_ixs.shape[0]

        # TODO: Do i need this? Probably yes
        # assert X.shape[0] == X_leaves_ixs.shape[0]

        # define the first array (starting point for the leaf outputs calculations)
        self.leaves_val_array_ = self.start_odds * np.ones(shape=(n_rows, 1))

        self.n_trees_ = X_leaves_ixs.shape[1]

        self.leaf_vals_map_ = {}
        # Loop over all the trees.
        for tree_index in range(self.n_trees_):
            # Define the prediction of the trees up to tree_index-1 (sum along axis 1)
            prev_proba = self.compute_probability(self.leaves_val_array_, tree_index=tree_index)

            # compute the gradient and hessian by using the predictions from the previous tree.
            g, h = self.f_loss_func(prev_proba, y)

            leaf_vals_ix, leaf_vals_map_ix = self._calculate_leaf_values(
                leaves_ixs=X_leaves_ixs, grad=g, hess=h, tree_index=tree_index, model_params=self.model_params
            )
            self.leaves_val_array_ = np.hstack([self.leaves_val_array_, leaf_vals_ix.reshape(-1, 1)])

            self.leaf_vals_map_[tree_index] = leaf_vals_map_ix

        return self

    def _predict_proba(self, X_leaves_ixs, tree_index=-1):

        X_leaves_ixs = assure_numpy_array(X_leaves_ixs)

        assert self.n_trees_ == X_leaves_ixs.shape[1]

        leaves_val_array = self._apply_leaf_map(X_leaves_ixs)

        if tree_index < 1:
            tree_index = self.n_trees_

        probs = self.compute_probability(leaves_val_array, tree_index=tree_index).reshape(-1, 1)

        return np.hstack([1 - probs, probs])

    def _apply_leaf_map(self, X_leaves_ixs):

        leaves_val_array = self.start_odds * np.ones(shape=(X_leaves_ixs.shape[0], 1))

        for tree_index in range(self.n_trees_):
            leaves = X_leaves_ixs[:, tree_index]

            # leaf_vals_ix = np.array([self.leaf_vals_map_[tree_index][ix] for ix in leaves])
            leaf_vals_ix = np.array([self._get_leaf_value(tree_index, ix) for ix in leaves])
            leaves_val_array = np.hstack([leaves_val_array, leaf_vals_ix.reshape(-1, 1)])

        return leaves_val_array

    def _get_leaf_value(self, tree_ix, leaf_ix):
        """It can happen that when fitting tboost, a specific leaf_ix is never found in the sample.

        For those cases, we should keep the leaf output to be 0
        Args:
            tree_ix: tree index
            leaf_ix: leaf index

        Returns: output of the leaf if leaf_index exists, otherwise 0
        """
        try:
            return self.leaf_vals_map_[tree_ix][leaf_ix]
        except KeyError:
            if self.verbosity > 0:
                warnings.warn(f"Never seen a leaf {leaf_ix} in {tree_ix}, assigning default value of 0")
            return 0

    @staticmethod
    def _calculate_leaf_values(leaves_ixs, grad, hess, tree_index, model_params):
        """Calculate the output of every leaf of the tree in position tree_index.

        Args:
            leaves_ixs (np.array): np.array, of the shape (n_rows, n_trees).
                Every entry corresponds to the index of the leaf of the model.
                This is returned by model.predict(X, return_leaf=True) for lgb, or model.apply(X)
                for xgboost models.
            grad (np.array): gradient array (for every row) computed from the previous predictions and target y.
            hess (np.array): hessian array (for every row) computed from the previous predictions and target y.
            tree_index (int): index of the decision tree
            model_params (dict): [description]
        """
        reg_lambda = model_params["reg_lambda"]
        learning_rate = model_params["learning_rate"]
        # learning_rate = 1

        leaf_vals_map = {}
        # Find the leaves indices for the tree ix
        leaves = leaves_ixs[:, tree_index]

        # Calculate the leaf values as per xgboost paper
        # eq.(5) in https://arxiv.org/pdf/1603.02754.pdf

        for leave_ix in np.unique(leaves):
            leaf_vals_map[leave_ix] = (
                learning_rate * grad[leaves == leave_ix].sum() / (hess[leaves == leave_ix].sum() + reg_lambda)
            )

        # Map every single leaf index to the new value
        leaf_values = np.array([leaf_vals_map[ix] for ix in leaves])
        return leaf_values, leaf_vals_map

    @staticmethod
    def compute_probability(leaves_val_array, tree_index=None):
        """Compute the probability given the outputs of the leaves.

        Args:
            leaves_val_array (np.array): two dimensional numpy array. expected shape( n_rows x n_trees+1).
                Contains the output of the leaves of the model. The first column must correspond to the
                "bias" term.
                Works with the output of recompute_leaves().
            tree_index (integer): index of the tree where to truncate the calculation of the probabilities.
                tree_index = 0 will return the probability of the base score (defaul 0.5 in Xgboost).
                tree_index = 1 returns the predicted probability of the 1st tree.

        Returns:
            np.array: array of predicted probabilities for class 1 at the tree_index^th tree
        """
        if tree_index is None:
            tree_index = leaves_val_array.shape[0]

        return _logistic(leaves_val_array[:, : (tree_index + 1)].sum(axis=1))
